// coded by Jeffrey Jestes
// jestes@cs.fsu.edu

// represents a prob string object

#include <vector>
#include <iostream>
#include <string>

using namespace std;


class instance {
   public:
      char c;
      double prob;
      bool original;
      instance() {
         c = 0;
         prob = 0;
         original = true;
      }
      instance( char c, double prob, bool original) {
         this->c = c;
         this->prob = prob;
         this->original = original;
      }
      instance( char c, double prob ) {
         this->c = c;
         this->prob = prob;
         this->original = true;
      }
      instance( const instance& i ){ 
         this->c = i.c;
         this->prob = i.prob;
         this->original = i.original;
      }
};


class random_var {
   public:
      std::vector<instance> v;
      random_var() {
         v.clear();
      };
      random_var( char c ) {
         v.push_back( instance( c, 1, true ) );
      }
      random_var( char c, double d ) {
         v.push_back( instance( c, d ) );
      }
      random_var( char c, bool b ) {
         v.push_back( instance( c, 1, b ) );
      }
      random_var( const random_var& r ) {
         v = r.v;
      }
      ~random_var() {
         this->v.clear();
      }

      bool contains( char c ) {
         for( unsigned int i = 0; i < v.size(); ++i ) {
            if( v[i].c == c ) return true;
         }
         return false;
      }

      random_var& operator=( const random_var& r ) {
         this->v.clear();
         for( unsigned int i = 0; i < r.v.size(); ++i ) {
            this->v.push_back( r.v[i] );
         }
         return *this;
      }
};

class prob_string {
   public:
      std::vector<random_var> ps;
      std::vector<random_var> aug_ps;

      prob_string() {}
      prob_string( const prob_string& p ) {
         ps = p.ps;
      }
      prob_string( char c ) {
         ps.push_back( random_var( c ) );
      }
      prob_string( prob_string& p ) {
         *this = p;
      }
      prob_string& operator=( const prob_string& p ) {
         for( unsigned int i = 0; i < ps.size(); ++i ) {
            ps[i].v.clear();
         }
         ps.clear();

         for( unsigned int i = 0; i < p.ps.size(); ++i ) {
            ps.push_back( p.ps[i] );
         }
         return *this;
      }

      void clear() {
         for( unsigned int i = 0; i < ps.size(); ++i ) {
            ps[i].v.clear();
         }
         ps.clear();
      }

      void print() {
         for( unsigned int i = 0; i < ps.size(); ++i ) {
            for( unsigned int j = 0; j < ps[i].v.size(); ++j ) {
               std::cout << ps[i].v[j].c << ' ' << ps[i].v[j].prob << ' ';
            }
            std::cout << "| ";
         }
         std::cout << std::endl;
      }

      void print_augps() {
         for( unsigned int i = 0; i < aug_ps.size(); ++i ) {
            for( unsigned int j = 0; j < aug_ps[i].v.size(); ++j ) {
               std::cout << aug_ps[i].v[j].c << ' ' << aug_ps[i].v[j].prob << ' ';
            }
            std::cout << "| ";
         }
         std::cout << std::endl;
      }

      void build_augps(int q) {
          aug_ps.resize(0);
          for( int i = 0; i < q - 1; ++i ) {
             aug_ps.push_back( random_var( '#' ) );
          }
          for( int i = 0; i < ps.size(); ++i ) {
             aug_ps.push_back( ps[i] );
          }
          for( int i = 0; i < q - 1; ++i ) {
             aug_ps.push_back( ';' );
          }
      }


      void printSQL() {
         for( unsigned int i = 0; i < ps.size(); ++i ) {
            for( unsigned int j = 0; j < ps[i].v.size(); ++j ) {
               if( ps[i].v.size() == 1 ) {
                  std::cout << ps[i].v[j].c;
               }
               else {
                  std::cout << ps[i].v[j].c << '#' << ps[i].v[j].prob << '#';
               }
            }
            std::cout << "|";
         }
         std::cout << "\r\n";
      }

      void printQGRAM( unsigned int key, unsigned int q, unsigned int string_length ) {

         std::vector< std::vector<char> > worlds;
         std::vector<double> prob;
         build_augps( q );

         for( unsigned int i = 0; enumerateQGRAM( worlds, prob, i, q ); ++i ) {
            for( unsigned int j = 0; j < worlds.size() ; ++j ) {
               std::cerr << key << '\t' << i << '\t';
               for( unsigned int k = 0; k < worlds[j].size(); ++k ) {
                  std::cerr << worlds[j][k];
               }
               std::cerr << "\t" << prob[j] << "\t" << string_length << "\r\n";
            }
            worlds.resize(0);
            prob.resize(0);
         }
      }


      void print( unsigned int i, unsigned int j) {
         if( i > ps.size() ) i = 0;
         for( ; i < ps.size() && i < j; ++i ) {
            for( unsigned int j = 0; j < ps[i].v.size(); ++j ) {
               std::cout << ps[i].v[j].c << ' ' << ps[i].v[j].prob << ' ';
            }
            std::cout << "| ";
         }
         std::cout << std::endl;
      }

      void enumerate( std::vector< std::vector<char> >& worlds, std::vector<double>& prob ) {
          worlds.push_back( std::vector<char>() );
          prob.push_back( 1 );

          unsigned int orig_size = 0;

          for( unsigned int i = 0; i < ps.size(); ++i ) {

             if( ps[i].v.size() == 1 ) {
                for( unsigned int j = 0; j < worlds.size(); ++j ) {
                   worlds[j].push_back( ps[i].v[0].c );
                   prob[j] *= ps[i].v[0].prob;
                }
             }
             else if( ps[i].v.size() > 1 ) {
                orig_size = worlds.size();
                
                // copy the original worlds by the number of choices in the current random var
                for( unsigned int j = 0; j < orig_size; ++j ) {
                   for( unsigned int k = 0; k < ps[i].v.size() - 1; ++k ) {
                      worlds.push_back( worlds[j] );
                      prob.push_back( prob[j] );
                   }
                }

                for( unsigned int j = 0; j < orig_size; ++j ) {
                   worlds[j].push_back( ps[i].v[0].c );
                   prob[j] *= ps[i].v[0].prob;
                }
                for( unsigned int j = orig_size; j < worlds.size(); ) {
                   for( unsigned int k = 1; k < ps[i].v.size(); ++k, ++j ) {
                      worlds[j].push_back( ps[i].v[k].c );
                      prob[j] *= ps[i].v[k].prob;
                   }
                }
             } // end else
          } // end for
       } // end enumerate

      void enumerate( std::vector< std::vector<char> >& worlds, std::vector<double>& prob, int index ) {
          worlds.push_back( std::vector<char>() );
          prob.push_back( 1 );

          unsigned int orig_size = 0;

          for( unsigned int i = 0; i < ps.size() && i < index; ++i ) {

             if( ps[i].v.size() == 1 ) {
                for( unsigned int j = 0; j < worlds.size(); ++j ) {
                   worlds[j].push_back( ps[i].v[0].c );
                   prob[j] *= ps[i].v[0].prob;
                }
             }
             else if( ps[i].v.size() > 1 ) {
                orig_size = worlds.size();
                
                // copy the original worlds by the number of choices in the current random var
                for( unsigned int j = 0; j < orig_size; ++j ) {
                   for( unsigned int k = 0; k < ps[i].v.size() - 1; ++k ) {
                      worlds.push_back( worlds[j] );
                      prob.push_back( prob[j] );
                   }
                }

                for( unsigned int j = 0; j < orig_size; ++j ) {
                   worlds[j].push_back( ps[i].v[0].c );
                   prob[j] *= ps[i].v[0].prob;
                }
                for( unsigned int j = orig_size; j < worlds.size(); ) {
                   for( unsigned int k = 1; k < ps[i].v.size(); ++k, ++j ) {
                      worlds[j].push_back( ps[i].v[k].c );
                      prob[j] *= ps[i].v[k].prob;
                   }
                }
             } // end else
          } // end for
       } // end enumerate


      bool enumerateQGRAM( std::vector< std::vector<char> >& worlds, std::vector<double>& prob, unsigned int pos, unsigned int q ) {

          if( pos + q > aug_ps.size() ) return false;

          worlds.push_back( std::vector<char>() );
          prob.push_back( 1 );

          unsigned int orig_size = 0;


          for( unsigned int i = pos; i < pos + q && i < aug_ps.size(); ++i ) {

             if( aug_ps[i].v.size() == 1 ) {
                for( unsigned int j = 0; j < worlds.size(); ++j ) {
                   worlds[j].push_back( aug_ps[i].v[0].c );
                   prob[j] *= aug_ps[i].v[0].prob;
                }
             }
             else if( aug_ps[i].v.size() > 1 ) {
                orig_size = worlds.size();
                
                // copy the original worlds by the number of choices in the current random var
                for( unsigned int j = 0; j < orig_size; ++j ) {
                   for( unsigned int k = 0; k < aug_ps[i].v.size() - 1; ++k ) {
                      worlds.push_back( worlds[j] );
                      prob.push_back( prob[j] );
                   }
                }

                for( unsigned int j = 0; j < orig_size; ++j ) {
                   worlds[j].push_back( aug_ps[i].v[0].c );
                   prob[j] *= aug_ps[i].v[0].prob;
                }
                for( unsigned int j = orig_size; j < worlds.size(); ) {
                   for( unsigned int k = 1; k < aug_ps[i].v.size(); ++k, ++j ) {
                      worlds[j].push_back( aug_ps[i].v[k].c );
                      prob[j] *= aug_ps[i].v[k].prob;
                   }
                }
             } // end else
          } // end for

          return true;
       } // end enumerate

       void QGRAMS( vector<char>& str, vector< vector<char> >& grams ) {
          
          grams.push_back( vector<char>() );
          grams.back().push_back( '#' );
          for( unsigned int i = 0; i < str.size(); ++i ) {
             grams.back().push_back( str[i] );
             grams.push_back( vector<char>() );
             grams.back().push_back( str[i] );
          }
          grams.back().push_back( ';' );
       }

       void QGRAMS3( string str, vector< vector<char> >& grams ) {

          str = "##" + str;
          str += ";;";
          
          for( unsigned int i = 0; i < str.size() - 2; ++i ) {
             grams.push_back( vector<char>() );
             grams.back().push_back( str[i] );
             grams.back().push_back( str[i+1] );
             grams.back().push_back( str[i+2] );
          }
       }

       void QGRAMS4( string str, vector< vector<char> >& grams ) {

          str = "###" + str;
          str += ";;;";
          
          for( unsigned int i = 0; i < str.size() - 3; ++i ) {
             grams.push_back( vector<char>() );
             grams.back().push_back( str[i] );
             grams.back().push_back( str[i+1] );
             grams.back().push_back( str[i+2] );
             grams.back().push_back( str[i+3] );
          }
       }

       void QGRAMS( string str, vector< vector<char> >& grams ) {
          
          grams.push_back( vector<char>() );
          grams.back().push_back( '$' );
          for( unsigned int i = 0; i < str.size(); ++i ) {
             grams.back().push_back( str[i] );
             grams.push_back( vector<char>() );
             grams.back().push_back( str[i] );
          }
          grams.back().push_back( '#' );
       }


       void print_enumeration( std::vector< std::vector<char> >& v, std::vector<double>& prob ) {
          for( unsigned int j = 0; j < v.size(); ++j ) {
             for( unsigned int k = 0; k < v[j].size(); ++k ) {
                cout << v[j][k];
             }
                cout << ' ' << prob[j] << endl;
          }
       }

       void print_enumeration( std::vector< std::vector<char> >& v, std::vector< vector< double > >& prob ) {
          for( unsigned int j = 0; j < v.size(); ++j ) {
             for( unsigned int k = 0; k < v[j].size(); ++k ) {
                cout << v[j][k] << ' ';
                cout << prob[j][k] << ' ';

             }
             cout << endl;
          }
       }
    
             
};
           
